(function ($, Drupal, drupalSettings) {

  Drupal.bpmn_io = {};

  Drupal.behaviors.bpmn_io = {
    attach: function (context, settings) {
      if (Drupal.bpmn_io.modeler === undefined) {
        window.addEventListener('resize', function () {
          let container = $('#bpmn-io');
          let offset = container.offset();
          let width = container.width();
          $('#bpmn-io .canvas')
            .css('top', offset.top)
            .css('left', offset.left)
            .css('width', width);
          $('#bpmn-io .property-panel')
            .css('max-height', $(window).height() - offset.top);
        }, false);
        window.dispatchEvent(new Event('resize'));
        Drupal.bpmn_io.addObserver('admin-toolbar');
        Drupal.bpmn_io.addObserver('gin_sidebar');

        Drupal.bpmn_io.isNew = settings.bpmn_io.is_new;
        Drupal.bpmn_io.modeler = window.modeler;
        Drupal.bpmn_io.layoutProcess = window.layoutProcess;
        Drupal.bpmn_io.expectToBeAccessible = window.expectToBeAccessible;
        Drupal.bpmn_io.expectToBeAccessible($('#bpmn-io .canvas'));
        Drupal.bpmn_io.eventBus = Drupal.bpmn_io.modeler.get('eventBus');
        Drupal.bpmn_io.loader = Drupal.bpmn_io.modeler.get('elementTemplatesLoader');
        Drupal.bpmn_io.loader.setTemplates(settings.bpmn_io.templates);
        Drupal.bpmn_io.elementTemplates = Drupal.bpmn_io.modeler.get('elementTemplates');
        Drupal.bpmn_io.templateChooser = Drupal.bpmn_io.modeler.get('elementTemplateChooser');
        Drupal.bpmn_io.open(settings.bpmn_io.bpmn);
        Drupal.bpmn_io.modeler.get('minimap').close();
        Drupal.bpmn_io.eventHandler();

        $('form input.button.modeler_api-save').click(function () {
          Drupal.bpmn_io.export();
          return false;
        });

        $('form input.button.modeler_api-export-svg').click(function () {
          Drupal.bpmn_io.saveSVG();
          return false;
        });
        $('form input.button.modeler_api-layout-process').click(function () {
          Drupal.bpmn_io.autoLayout();
          return false;
        });
        $('form input.button.modeler_api-close').click(function () {
          window.location = drupalSettings.modeler_api.collection_url;
          return false;
        });
        if (drupalSettings.modeler_api.useFormApi) {
          $('#bpmn-io .property-panel').hide();
        }
        else {
          Drupal.bpmn_io.dragAndDrop($('#bpmn-io .property-panel.in-canvas')[0]);
        }

        $('form.ai-agent-form').submit( function(eventObj) {
          Drupal.bpmn_io.addDataToFormField();
          return true;
        });
      }
      Drupal.bpmn_io.addObserver('drupal-off-canvas-wrapper');
      Drupal.bpmn_io.prepareMessages();
    },
  };

  Drupal.bpmn_io.addObserver = function (id) {
    let element = document.getElementById(id);
    if (element === null) {
      return;
    }
    new MutationObserver(function (mutations) {
      mutations.forEach(function(mutation) {
        window.dispatchEvent(new Event('resize'));
      });
    }).observe($('body')[0], {
      attributes: true
    });
    new ResizeObserver(function () {
      setTimeout(function () {
        window.dispatchEvent(new Event('resize'));
      }, 200);
    }).observe(element);
  };

  Drupal.bpmn_io.prepareMessages = function () {
    $('.messages-list:not(.bpmn-io-processed)')
      .addClass('bpmn-io-processed')
      .click(function () {
        $(this).empty();
      });
  };

  Drupal.bpmn_io.addDataToFormField = async function () {
    let result = await Drupal.bpmn_io.modeler.saveXML({ format: true });
    $('input[name="modeler_api_data"]')[0].value = result.xml;
    $('input#bio-properties-panel-id').remove();
  };

  Drupal.bpmn_io.export = async function () {
    $('.messages-list').empty();
    let result = await Drupal.bpmn_io.modeler.saveXML({ format: true });
    if (drupalSettings.modeler_api.token_url === undefined) {
      return;
    }
    let response = await fetch(drupalSettings.modeler_api.token_url);
    if (response.ok) {
      let token = await response.text();
      let request = Drupal.ajax({
        url: drupalSettings.modeler_api.save_url,
        submit: result.xml,
        beforeSend: function (xhr) {
          xhr.overrideMimeType("application/json;charset=UTF-8");
          xhr.setRequestHeader("X-CSRF-Token", token);
        },
        progress: {
          type: 'fullscreen',
          message: Drupal.t('Saving model...'),
        },
      });
      request.execute();
    }
    Drupal.bpmn_io.prepareMessages();
  };

  Drupal.bpmn_io.open = async function (bpmnXML) {
    await Drupal.bpmn_io.modeler.importXML(bpmnXML);
    Drupal.bpmn_io.canvas = Drupal.bpmn_io.modeler.get('canvas');
    Drupal.bpmn_io.overlays = Drupal.bpmn_io.modeler.get('overlays');
    Drupal.bpmn_io.canvas.zoom('fit-viewport');
    Drupal.bpmn_io.updateCanvas();
  };

  Drupal.bpmn_io.dragAndDrop = function (panel) {
    if (panel === undefined) {
      return;
    }
    const BORDER_SIZE = 4;
    let m_pos;

    function resize(e) {
      const dx = m_pos - e.x;
      m_pos = e.x;
      panel.style.width = (parseInt($(panel).outerWidth()) - BORDER_SIZE + dx) + 'px';
    }

    panel.addEventListener('mousedown', function (e) {
      if (e.offsetX < BORDER_SIZE) {
        m_pos = e.x;
        document.addEventListener('mousemove', resize, false);
      }
    }, false);
    document.addEventListener('mouseup', function () {
      document.removeEventListener('mousemove', resize, false);
    }, false);
  };

  Drupal.bpmn_io.saveSVG = function () {
    Drupal.bpmn_io.canvas.focus();
    Drupal.bpmn_io.modeler.saveSVG({ format: true }).then((model) => {
      let svgBlob = new Blob([model.svg], {
        type: 'image/svg+xml'
      });
      let downloadLink = document.createElement('a');
      downloadLink.download = drupalSettings.bpmn_io.id + '.svg';
      downloadLink.innerHTML = 'Get BPMN SVG';
      downloadLink.href = window.URL.createObjectURL(svgBlob);
      downloadLink.onclick = function (event) {
        document.body.removeChild(event.target);
      };
      downloadLink.style.visibility = 'hidden';
      document.body.appendChild(downloadLink);
      downloadLink.click();
    });
  };

  Drupal.bpmn_io.autoLayout = async function () {
    // Export and auto-layout the model.
    const result = await Drupal.bpmn_io.modeler.saveXML({ format: true });
    const diagramWithLayoutXML = await Drupal.bpmn_io.layoutProcess(result.xml);
    Drupal.bpmn_io.open(diagramWithLayoutXML);
  };

  Drupal.bpmn_io.updateCanvas = function (element) {
    let idField = $('#bio-properties-panel-id');
    if (!Drupal.bpmn_io.isNew) {
      if (element === undefined) {
        Drupal.bpmn_io.processId = $(idField)[0].value;
        $(idField)
          .hide()
          .parent('.bio-properties-panel-textfield').find('label').append('<span>: ' + Drupal.bpmn_io.processId + '</span>');
      }
      else if (element.id === Drupal.bpmn_io.processId) {
        if (!$(idField).is(':hidden')) {
          $(idField)
            .hide()
            .parent('.bio-properties-panel-textfield').find('label').append('<span>: ' + Drupal.bpmn_io.processId + '</span>');
        }
      }
    }
  };

  Drupal.bpmn_io.findPluginId = function (element) {
    let pluginId = undefined;
    if (element.di.bpmnElement.extensionElements !== undefined) {
      element.di.bpmnElement.extensionElements.values.forEach(function (value) {
        if (value.values !== undefined && value.values[0].name === 'pluginid') {
          pluginId = value.values[0].value;
          return true;
        }
      });
    }
    return pluginId;
  };

  Drupal.bpmn_io.loadConfigForm = function (element, pluginId) {
    $(document).one('ajaxComplete', function handleAjaxComplete() {
      if (element.di.bpmnElement.name === undefined) {
        let name = $('.ui-dialog-titlebar h2').text();
        Drupal.bpmn_io.modeler.get('modeling').updateProperties(element, {name: name});
      }
      if (element.di.bpmnElement.extensionElements !== undefined) {
        let alwaysUpdate = [];
        element.di.bpmnElement.extensionElements.values.forEach(function (value) {
          if (value.name !== undefined) {
            $('#drupal-off-canvas form [name="' + value.name + '"]')
                .each(function () {
                  if ($(this).hasClass('machine-name-target')) {
                    alwaysUpdate.push(this);
                  }
                  if ($(this).attr('type') === 'checkbox') {
                    $(this)
                        .prop('checked', value.string === 'yes')
                        .change();
                  }
                  else {
                    $(this)
                        .val(value.string)
                        .change();
                  }
                })
                .on('change', function () {
                  let updateField = function (currentField) {
                    let fieldName = $(currentField).attr('name');
                    let newValue = ($(currentField).attr('type') === 'checkbox') ?
                        ($(currentField).prop('checked') ? 'yes' : 'no') :
                        $(currentField).val();
                    let i = 0;
                    element.di.bpmnElement.extensionElements.values.forEach(function (objvalue) {
                      if (objvalue.name !== undefined && objvalue.name === fieldName) {
                        element.di.bpmnElement.extensionElements.values[i].string = newValue;
                        return true;
                      }
                      i++;
                    });
                  };
                  updateField(this);
                  alwaysUpdate.forEach(function (item) {
                    updateField(item);
                  });
                });
          }
        });
      }
    });

    let offCanvas = document.getElementById('drupal-off-canvas-wrapper');
    if (offCanvas) {
      let style = offCanvas.currentStyle || window.getComputedStyle(offCanvas);
      Drupal.bpmn_io.offCanvasWidth = parseFloat(style.width);
    }
    let config = {};
    element.di.bpmnElement.extensionElements.values.forEach(function (value) {
      if (value.name !== undefined) {
        config[value.name] = value.string;
      }
    });

    let request = Drupal.ajax({
      url: drupalSettings.modeler_api.config_url,
      submit: {
        entityId: Drupal.bpmn_io.processId,
        isNew: Drupal.bpmn_io.isNew,
        type: element.type,
        pluginId: pluginId,
        config: config,
        width: Drupal.bpmn_io.offCanvasWidth || '',
      },
      progress: {
        type: 'fullscreen',
        message: Drupal.t('Loading config form...'),
      },
    });
    request.execute();
  };

  Drupal.bpmn_io.closeOffCanvas = function () {
    $('#drupal-off-canvas-wrapper .ui-dialog-titlebar-close').click();
  };

  Drupal.bpmn_io.elementSelected = function (element) {
    let templates = Drupal.bpmn_io.elementTemplates.getAll(element);
    if (templates.length === 0) {
      Drupal.bpmn_io.closeOffCanvas();
      return;
    }
    let pluginId = Drupal.bpmn_io.findPluginId(element);
    if (pluginId === undefined) {
      if (templates.length > 1) {
        Drupal.bpmn_io.eventBus.once('elementTemplateChooser.chosen', (e) => {
          Drupal.bpmn_io.eventBus.fire('popupMenu.close');
          Drupal.bpmn_io.elementTemplates.applyTemplate(element, e.template);
          pluginId = Drupal.bpmn_io.findPluginId(element);
          if (pluginId !== undefined) {
            Drupal.bpmn_io.loadConfigForm(element, pluginId);
          }
        });
        Drupal.bpmn_io.eventBus.fire('elementTemplates.select', {element: element});
        return true;
      }
      Drupal.bpmn_io.elementTemplates.applyTemplate(element, templates[0]);
      pluginId = Drupal.bpmn_io.findPluginId(element);
    }
    Drupal.bpmn_io.loadConfigForm(element, pluginId);
  };

  Drupal.bpmn_io.eventHandler = function () {
    if (drupalSettings.modeler_api.useFormApi) {
      Drupal.bpmn_io.eventBus.on('selection.changed', 250 , function (e) {
        if (e.newSelection[0] === undefined) {
          Drupal.bpmn_io.closeOffCanvas();
        }
        else if (e.newSelection[1] === undefined) {
          Drupal.bpmn_io.elementSelected(e.newSelection[0]);
        }
      });
    }
    else {
      Drupal.bpmn_io.eventBus.on('element.click', function (e) {
        setTimeout(function () {
          Drupal.bpmn_io.updateCanvas(e.element);
        }, 5);
      });
    }
  };

})(jQuery, Drupal, drupalSettings);
